/**********************************************************************
 * File Name: CommChannel.java
 *
 * Copyright (C) 2000, US Air Force Research Laboratory,
 * Rome Research Site, Rome NY 13424, USA.
 *
 * Use of this software implies acceptance of the terms described in
 * accompanying license.txt file.
 *
 *
 * Authors: Jennifer Dzek, James Lawton, John Crowter,
 *          Bob Paragi, and Joe Beasock
 *
 * Creation Date: July 1999
 *
 * Description:
 *
 * Modification History:
 *
 **********************************************************************/

import java.lang.*;
import java.io.*;
import java.util.*;

public class CommChannel extends SimObject {
    private ArrayList sensors;
    private ArrayList messages;
    // private boolean send_to_cc;

    // Transmission speed in characters per millisecond:
    public static final float TRANSMISSION_SPEED = 1.92f;

    // Adjustment factor to account for the fact we aren't as 
    // efficient as the hardware in packing messages.
    public static final float ADJ_FACTOR = 2.0f;

    // Event Types:
    private final int DELIVER_MESSAGE_EVENT = 1;


    public CommChannel (String newname, SimEnvironment simenv) {
	sensors = new ArrayList();
	messages = new ArrayList();
	// send_to_cc = false;
	name = newname;
	env = simenv;
    }

    // Add SNSR to the list of Sensors listening on this channel
    public void register (Sensor snsr) {
	if (! sensors.contains(snsr) ) {
	    sensors.add(snsr);
	}
    }

    // Remove SNSR to the list of Sensors listening on this channel
    public void unregister (Sensor snsr) {
	int ind = sensors.indexOf(snsr);
	if (ind >= 0)
	    sensors.remove(ind);
    }


    // Compute the number of clock ticks to transmit this message.
    // Crude at this point, but effective.
    private int computeMessageDuration(Message msg) {
	float mdur = msg.text.length() * TRANSMISSION_SPEED;
	int ticks = Math.round( (mdur / (float) env.clock.ticksize) / ADJ_FACTOR);
	//System.out.println("Message length = " + msg.text.length() +
	//		   ", ticks = " + ticks);
	return(ticks);
    }
    
    // Add the given message to the message queue, and schedule a
    // message delivery event if needed.
    public void sendMessage (Message msg) {
	if (env.Debug == -1) 
	    System.out.println(name + "(" + env.clock.simtime + 
			       "): Adding new message [" +  msg.message_id() + 
			       ", "  + msg.sender() + " --> " + msg.receiver() + "]");

	// Add the message to the message queue
	CommMessage newMsg = new CommMessage(msg.message_id, computeMessageDuration(msg));
	newMsg.messages.add(msg);
	messages.add(newMsg);

	// Schedule the next message delivery event, if needed
	scheduleDeliveryEvent();
    }

    public void sendMessage (CommMessage msgs) {
	if (env.Debug == -1) 
	    System.out.println(name + "(" + env.clock.simtime + 
			       "): Adding new multicast message:" +  msgs.message_id);

	// Add the message to the message queue
	Message msg = (Message) msgs.messages.get(0);
	msgs.duration = computeMessageDuration(msg);
	messages.add(msgs);

	// Schedule the next message delivery event, if needed
	scheduleDeliveryEvent();
    }


    // Schedule the next Message Delivery Event, if none are 
    // currently scheduled.  If one is, it will schedule the next one.
    private void scheduleDeliveryEvent () {
	if ( event_list.isEmpty() ) {
	    CommMessage nextMsg = (CommMessage) messages.get(0);
	    int evtime = nextMsg.duration + env.clock.simtime;
	    if (env.Debug > 1) 
		System.out.println("  Scheduling delivery event for simtime " + evtime);
	    schedule(DELIVER_MESSAGE_EVENT, evtime);
	}
    }

    // Event Dispatching
    public void dispatchEvent (EventStruct ev) {
	int etype = ev.event_type;
	int simtime = ev.event_time;

	if (env.Debug > 2)
	    System.out.println("CommChannel " + name + " dispatching event type: " 
			       + etype + " time: " + simtime);

	switch (etype) {
	case DELIVER_MESSAGE_EVENT:
	    deliverMessage(simtime);
	    break;
	}
    }

    private void deliverMessage(int simtime) {
	CommMessage msg = (CommMessage) messages.remove(0);

	// Multicast deliver the message(s) at the head of the 
	// message queue to all registered sensors.
	for (ListIterator snsrs = sensors.listIterator(); snsrs.hasNext(); ) {
	    Sensor snsr = (Sensor) snsrs.next();
	    for (ListIterator msgs = msg.messages.listIterator(); msgs.hasNext(); ) {
		Message nextMsg = (Message) msgs.next();
		
		if (env.Debug == -1)
		    System.out.println(name + "(" + env.clock.simtime + 
				       "): Delivering message [" +  nextMsg.message_id() + 
				       ", "  + nextMsg.sender() + " --> " + nextMsg.receiver() + "]");
		
		snsr.receiveMessage(nextMsg);
	    }
	}
	
	//Schedule the next event
	if (! messages.isEmpty() ) {
	    msg = (CommMessage) messages.get(0);
	    schedule(DELIVER_MESSAGE_EVENT, simtime + msg.duration);
	}
    }
}
